package org.polymodel.algebra.tom;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.polymodel.algebra.ComparisonOperator;
import org.polymodel.algebra.IntConstraint;
import org.polymodel.algebra.IntConstraintSystem;
import org.polymodel.algebra.IntExpression;
import org.polymodel.algebra.IntTerm;
import org.polymodel.algebra.Variable;
import org.polymodel.algebra.affine.AffineExpression;
import org.polymodel.algebra.affine.AffineTerm;
import org.polymodel.algebra.polynomials.PolynomialExpression;
import org.polymodel.algebra.quasiAffine.NestedQuasiAffineTerm;
import org.polymodel.algebra.quasiAffine.QuasiAffineExpression;
import org.polymodel.algebra.quasiAffine.QuasiAffineOperator;
import org.polymodel.algebra.quasiAffine.QuasiAffineTerm;
import org.polymodel.algebra.quasiAffine.SimpleQuasiAffineTerm;
import org.polymodel.algebra.reductions.ReductionOperator;

@SuppressWarnings("all")
public class ArithmeticOperations  {

	%include { sl.tom }
	%include { long.tom}     

	%include { algebra_common.tom}   
	%include { algebra_terms.tom}     
	%include { algebra_all_ops.tom}          
    
	public static boolean VERBOSE=false;  
	public static long ZERO=0; 
    
	public static int Math_max(int a, int b) {
		return Math.max(a,b);
	}  

	public static long Math_max(long a, long b) {
		return Math.max(a,b);
	}  

	public static int Math_min(int a, int b) {
		return Math.min(a,b);
	} 

	public static long Math_min(long a, long b) {
		return Math.min(a,b);
	}

	public static int Math_mod(int a, int b) {
		return a%b;
	} 
	
	public static int Math_mul(int a, int b) {
		return a*b;
	} 

	public static int Math_floord(int a, int b) {
		return a/b;
	} 

	public static int Math_ceild(int a, int b) { 
		return (a+b-1)/b;
	} 
 
 

	public static IntTerm negate(IntTerm a) {
		return scale(a,-1);
	}  

	public static IntExpression negate(IntExpression a) {
		return scale(a,-1);
	}

	public static AffineExpression negate(AffineExpression a) {
		return ((AffineExpression) negate(((IntExpression)a)));
	}


  
	public static IntTerm scale(IntTerm a, long value) {
		%match( a ) {
			term(c,v) -> {
				return `term((value)*(c),v);
			} 
			x@mul(e,coef) -> {
				return `mul(e.copy(),(value)*(coef));
			}
			x@mod(e,coef) -> {
				if((value % `coef)==0) {
					return `mod(e.copy(),coef);
				} else {
					return `nmul(qaffine(terms(mod(e.copy(),coef))),value);
				}
			}
			x@floor(e,coef) -> {
				return `nmul(qaffine(terms(floor(e.copy(),coef))),value);
			}

			x@div(e,coef) -> {  
				return `nmul(qaffine(terms(floor(e.copy(),coef))),value);
			}
			nqterm(e) -> {
				return `nqterm(scale(e,value));
			}
		}
		throw new RuntimeException("scaleTerm("+a+","+value+") not yet implemented");
	}  

	public static AffineExpression scale(AffineExpression a, long value) {
		return ((AffineExpression) scale(((IntExpression)a),value));
	}

	public static IntExpression scale(IntExpression a, long value) {
		IntExpression _a = a.copy();
		%match(_a) {
			x@affine(terms(x1*)) -> {
				AffineExpression r = (AffineExpression)`affine(terms());
				for(IntTerm t : `x1) {
					r.getTerms().add((AffineTerm)scale(t,value)); 
				}
				return r.simplify();
			}
			x@qaffine(terms(x1*)) -> {
				QuasiAffineExpression r = (QuasiAffineExpression)`qaffine(terms());
				for(QuasiAffineTerm t : ((QuasiAffineExpression)`x).getTerms()) {
					r.getTerms().add((QuasiAffineTerm) scale(t,value));
				}
				return r.simplify(); 
			}
		}
		throw new RuntimeException("scale("+a+","+value+") not yet implemented");
	}

	public static AffineExpression add(AffineExpression a,AffineExpression b) {
		return ((AffineExpression)add((IntExpression)a,(IntExpression)b)); 
	}
	public static IntExpression add(IntExpression a,IntExpression b) {
		IntExpression _a = a.copy();
		IntExpression _b= b.copy() ;
		%match(_a,_b) {
			x@affine(terms(x1*)),y@affine(terms(x2*)) -> {
				return `affine(terms(x1*,x2*));
			}
			x@affine(terms(x1*)),y@qaffine(terms(x2*)) -> {
				return `qaffine(terms(mul(x,1), x2));
			}
			y@qaffine(terms(x2*)),x@affine(terms(x1*)) -> {
				return `qaffine(terms(mul(x,1), x2));
			}
			y@qaffine(terms(x2*)),x@qaffine(terms(x1*)) -> {
				return `qaffine(terms(mul(x,1), x2));
			}
		}
		throw new RuntimeException("add("+a+") not yet implemented");
	}   

	public static IntExpression add(long a,IntExpression b) {
		return add(b,`affine(terms(constant(a))));
	}

	public static IntExpression add(IntExpression b,long a) {
		return add(b,`affine(terms(constant(a))));
	}

	public static IntExpression sub(IntExpression a,IntExpression b) {
		IntExpression _a = a.copy();
		IntExpression _b= b.copy() ;
		%match(_a,_b) {
			x@affine(terms(x1*)),y@affine(terms(x2*)) -> {
				return add(`x, negate(`y));
			}
			x@affine(terms(x1*)),y@qaffine(terms(x2*)) -> {
				return add(`x, negate(`y));
			}
			y@qaffine(terms(x1*)),x@affine(terms(x2*)) -> {
				return add(`x, negate(`y));
			}
			y@qaffine(terms(x1*)),x@qaffine(terms(x2*)) -> {
				return add(`x, negate(`y));
			}
		}
		throw new RuntimeException("sub("+a+") not yet implemented");
	} 



	public static IntExpression debug(String name, IntExpression input, IntExpression output) {
		if(VERBOSE) {
			System.out.println("rule "+name+" :\n\t"+input+" -> "+output);
		}
		return output;
	}

}