package org.polymodel.algebra.tom;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.polymodel.algebra.ComparisonOperator;
import org.polymodel.algebra.IntConstraint;
import org.polymodel.algebra.IntConstraintSystem;
import org.polymodel.algebra.IntExpression;
import org.polymodel.algebra.IntTerm;
import org.polymodel.algebra.Variable;
import org.polymodel.algebra.affine.AffineExpression;
import org.polymodel.algebra.affine.AffineTerm;
import org.polymodel.algebra.internal.AlgebraIntrospector;
import org.polymodel.algebra.polynomials.PolynomialExpression;
import org.polymodel.algebra.quasiAffine.QuasiAffineOperator;
import org.polymodel.algebra.reductions.ReductionOperator;

/**
 * Variable isolation tool.
 * 
 * @author antoine 
 * 
 */
@SuppressWarnings("all")
public class Isolate{ 

	%include { sl.tom }

	%include { algebra_common.tom}   
	%include { algebra_terms.tom}    
	%include { algebra_all_ops.tom} 
 
	private static boolean DEBUG=false;

	private static void debug(String mess) {
		if(DEBUG) {
			System.out.println(mess);
		} 
	}
	/**
	 * Build an expression corresponding to a variable isolation. Variable is removed and
	 * other terms are negated. Input expression isn't modified.
	 * 
	 * @param expression 
	 * @param var variable to isolate
	 * @return a new expression
	 */
	public static IntExpression isolate(IntExpression expression, Variable var) {
		try {
			return `OnceTopDown(VariableIsolation(var)).visitLight(expression,new AlgebraIntrospector()); 
		} catch(Exception e) { 
			e.printStackTrace(); 
			throw new RuntimeException("Visitor failure on "+expression+ ":"+e.getMessage());
		}
	}
  
	%strategy VariableIsolation(var:V) extends Identity() {
		visit E {
			affine(terms(x1*,b@term(k,var),x2*)) -> {
				if(`var==var) {
					AffineExpression res= ArithmeticOperations.scale((AffineExpression)`affine(terms(x1*,x2*)),-1);
					if (`k==1 || `k==-1) {					
						return `res; 
					}
					else
						return `qaffine(terms(floor(res,k)));
				}
			}
			a -> {
					debug("Fail on "+`a+" with " +`var+" for "+var);`
					throw new RuntimeException("Fail in Isolate("+`a+") for "+var);  
			}
		} 
	}
}