package org.polymodel.algebra.tom;
 
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.polymodel.algebra.ComparisonOperator;
import org.polymodel.algebra.IntConstraint;
import org.polymodel.algebra.IntConstraintSystem;
import org.polymodel.algebra.IntExpression;
import org.polymodel.algebra.IntTerm;
import org.polymodel.algebra.Variable;
import org.polymodel.algebra.affine.AffineExpression;
import org.polymodel.algebra.affine.AffineTerm;
import org.polymodel.algebra.internal.AlgebraIntrospector;
import org.polymodel.algebra.polynomials.PolynomialExpression;
import org.polymodel.algebra.quasiAffine.NestedQuasiAffineTerm;
import org.polymodel.algebra.quasiAffine.QuasiAffineExpression;
import org.polymodel.algebra.quasiAffine.QuasiAffineOperator;
import org.polymodel.algebra.quasiAffine.QuasiAffineTerm;
import org.polymodel.algebra.quasiAffine.SimpleQuasiAffineTerm;
import org.polymodel.algebra.reductions.ReductionExpression;
import org.polymodel.algebra.reductions.ReductionOperator;

@SuppressWarnings("all")
public class SimplifyIntExpression  {

	%include { sl.tom }  
	%include { long.tom}     
	%include { algebra_common.tom}   
	%include { algebra_terms.tom}    
	%include { algebra_all_ops.tom}      

	public static int Math_max(int a, int b) {
		return Math.max(a,b);
	}

	public static long Math_max(long a, long b) {
		return Math.max(a,b);
	}  

	public static int Math_min(int a, int b) {
		return Math.min(a,b);
	} 

	public static long Math_min(long a, long b) {
		return Math.min(a,b);
	}

	public static int Math_mod(int a, int b) {
		return a%b;
	} 
	
	public static int Math_mul(int a, int b) {
		return a*b;
	} 

	public static int Math_floord(int a, int b) {
		return a/b;
	} 

	public static int Math_ceild(int a, int b) { 
		return (a+b-1)/b;
	} 
 
	public static IntExpression simplify(IntExpression expression) {
		try {
			if(VERBOSE) {  
				System.out.println("*** Simplifying "+expression);
			} 
			IntExpression res = `Innermost(Simplify()).visitLight(expression,new AlgebraIntrospector());
			if(VERBOSE) {
				System.out.println("*** into "+res+"\n");
			}
			return res;  
		} catch(Exception e) { 
			System.err.println("Visitor failure on "+expression+ ":"+e.getMessage());
			throw new RuntimeException("Visitor failure on "+expression+ ":"+e.getMessage(),e);
			//e.printStackTrace(); 
		}
	} 
 

	public static IntExpression debug(String name, IntExpression input, IntExpression output) {
		if(VERBOSE) {
			System.out.println("rule "+name+" :\n\t"+input+" -> "+output);
		}
		return output;
	}
	public static boolean VERBOSE=false;   
	public static long ZERO=0;
	
	%strategy Simplify() extends Fail() {
		visit E {  
			in@affine(terms()) -> {
				return debug("R-1",`in,`affine(terms(constant(0)))); }
			in@affine(terms(x1*,constant(a),x2*,constant(b),x3*))  -> { 
				return debug("R0",`in,`affine(terms(x1*,x2*,x3*,constant(a+b)))); 
			}
			in@affine(terms(x1,constant(0L),x2*)) 	->  
				{ return debug("R1a",`in,`affine(terms(x1,x2*))); }
			in@affine(terms(x1*,constant(0L),x2)) 	->  
				{ return debug("R1b",`in,`affine(terms(x1*,x2))); }
			in@affine(terms(x1,term(0L,v),x2*)) -> 
				{ return debug("R2a",`in,`affine(terms(x1,x2*))); }
			in@affine(terms(x1*,term(0L,v),x2)) -> 
				{ return debug("R2b",`in,`affine(terms(x1*,x2))); }
			in@affine(terms(x1*,term(a,v),x2*,term(b,v),x3*)) -> 
				{ 
					long d=(`a)+(`b);
					if(d!=0) {
						return debug("R3",`in,`affine(terms(term(a+b,v),x1*,x2*,x3*)));
					} else {
						return debug("R3",`in,`affine(terms(x1*,x2*,x3*)));
					}
				 }
 			in@sum(EL(x1*,affine(terms(a*)),x2*,affine(terms(b*)),x3*)) -> 
 				{ return debug("R4",`in,`sum(EL(affine(terms(a,b)),x1*,x2*,x3*))); }
 				
			in@sum(EL(x1*,affine(terms()),x2*)) 			-> 
				{ return debug("R5",`in,`sum(EL(x1*,x2*))); }
			in@sum(EL()) -> 
				{ return debug("R6",`in,`affine(terms(constant(0L)))); }
			in@sum(EL(a)) -> 
				{ return debug("R7",`in,`a); }
			in@sum(EL(a*,sum(EL(b*)),c*)) -> 
				{ return debug("R8",`in,`sum(EL(a,b,c))); }
			in@sum(EL(x1*,affine(terms(b*)),x2*,affine(terms(c*)),x3*)) -> 
				{ return debug("R20",`in,`sum(EL(x1*,x2*,x3*,affine(terms(b*,c*))))); }
			in@sum(EL(x1*,qaffine(terms(b*)),x2*,qaffine(terms(c*)),x3*)) -> 
				{ return debug("R21",`in,`sum(EL(x1*,x2*,x3*,qaffine(terms(b*,c*))))); }

 
			 //overlaps with other rules defined for min/max below
			//in@min(EL(a)) -> 
			//	{ return debug("R9",`in,`a); }
			//in@min(EL(a*,min(EL(b*)),c*)) -> 
			//	{ return debug("R10",`in,`min(EL(a,b,c))); }
			
			//in@max(EL(a)) -> 
			//	{ return debug("R11",`in,`a); }
			//in@max(EL(a*,min(EL(b*)),c*)) -> 
			//	{ return debug("R12",`in,`max(EL(a,b,c))); }

 

			in@qaffine(terms(mul(aff@affine(_),a)))  -> {		
				return debug("R13",`in,ArithmeticOperations.scale((AffineExpression)`aff,`a)); } 

			in@qaffine(terms(mul(qaff@qaffine(terms(x*)),1L)))  -> {		
				return debug("R13'",`in,`qaffine(terms(x*))); } 

			in@qaffine(terms(nmul(qaff@qaffine(terms(x*)),1L)))  -> {		
				return debug("R13'",`in,`qaffine(terms(x*))); } 


		// (a)%b =(a%b)
			in@qaffine(terms(x1*,op@(floor|mod)(affine(terms(constant(a))),b),x2*))  -> {		
						long modval= 0;
						switch (((QuasiAffineTerm)`op).getOperator()) {
							case MOD:
								modval= (`a)%(`b); 
								break;  
							case FLOOR:
								modval= (`a)/(`b);
								break;
							case CEIL:
								modval= (`a + `b -1)%(`b);
								break;
							default: 
								modval= (`a)%(`b);
								break; 
						}
						long one =  1; 
						return debug("R14",`in,`qaffine(terms(x1*,mul(affine(terms(constant(modval))),one),x2*)));
			} 


			in@qaffine(terms(x1*,mul(qaffine(terms(mul(e,a))),b),x2*))  -> {
				long d = (`a)*(`b);
				if(d!=0) {
					return debug("R15prime",`in,`qaffine(terms(x1*,x2*,mul(e,d))));
				} else {
					return debug("R15prime",`in,`qaffine(terms(x1*,x2*)));
				}
			} 


			in@qaffine(terms(mul(qaffine(terms(mul(e@affine(terms(_*)),a))),b)))  -> {
				long d = (`a)*(`b);
				if(d!=0) {
					return debug("R15prime",`in,`qaffine(terms(mul(e,d))));
				} else {  
					return debug("R15prime",`in,`affine(terms()));
				}
			} 
 
			in@qaffine(terms(mul(qaffine(terms(mul(e@qaffine(terms(_*)),a))),b)))  -> {
				long d = (`a)*(`b);
				if(d!=0) {
					return debug("R15prime",`in,`qaffine(terms(mul(e,d))));
				} else {  
					return debug("R15prime",`in,`affine(terms()));
				}
			} 

			in@qaffine(terms(x1*,(floor|mod|mul)(affine(terms(x3*)),0L),x2*))  -> {		
				return debug("R15",`in,`qaffine(terms(x1*,x2*)));
			} 

			in@qaffine(terms(x1*,(floor|mod|mul)(affine(terms()),b),x2*))  -> {		
				return debug("R15",`in,`qaffine(terms(x1*,x2*)));
			} 

			// (bx)%b =0
			in@qaffine(terms(x1*,mod(affine(terms(x2*,term(b,a),x3*)),b),x4*))  -> {		
				return debug("R16",`in,`qaffine(terms(x1*,mod(affine(terms(x2*,x3*)),b),x4*)));
			} 
 
			// (bx)%b =0
			in@qaffine(terms(x1*,mod(affine(terms(x2*,constant(b),x3*)),b),x4*))  -> {		
				return debug("R16bis",`in,`qaffine(terms(x1*,mod(affine(terms(x2*,x3*)),b),x4*)));
			} 

			// (bx)/b =x
			in@qaffine(terms(x1*,floor(affine(terms(x2*,term(b,a),x3*)),b),x4*))  -> {		
				return debug("R17",`in,`qaffine(terms(x1*,mul(affine(terms(term(1,a))),1),floor(affine(terms(x2*,x3*)),b),x4*)));
			} 

			// ax*ay = a*(x+y)
			in@qaffine(terms(x1*,mul(a,c),x2*,mul(b,c),x3*))  -> {		
				AffineExpression res=  ArithmeticOperations.add((AffineExpression)`a,(AffineExpression)`b);
				return debug("R22",`in,`qaffine(terms(x1*,x2*,x3,mul(res,c))));
			}   

			in@qaffine(terms(x1*,floor(a,c),x2*,floor(b,c),x3*))  -> {		
				AffineExpression res=  ArithmeticOperations.add((AffineExpression)`a,(AffineExpression)`b);
				return debug("R23",`in,`qaffine(terms(x1*,x2*,x3,floor(res,c))));
			} 
 
			in@qaffine(terms(
							x1*,
							mul(e0@affine(terms(x*)),a),
							x2*,
							mul(e1@affine(terms(y*)),d),
							x3*
						)
					)  -> {		  
						AffineExpression res = 
						ArithmeticOperations.add(
							ArithmeticOperations.scale((AffineExpression)`e0,`a), 
							ArithmeticOperations.scale((AffineExpression)`e1,`d)
						);
						return debug("R18",`in,`qaffine(terms(x1*,x2*,x3*,mul(res,1))));
					

					} 


			in@reduction(EL(x),b) -> {
				return debug("red(E)->E", `in, `x); 
			}
			in@sum(EL(x1*,affine(terms(a*)),x2*,affine(terms(b*)),x3*)) -> 
											{ return debug("sum0", `in, `sum(EL(affine(terms(a*,b*)),x1*,x2*,x3*))); }
			in@sum(EL(affine(a))) -> 		{ return debug("sum1", `in, `affine(a)); }
			in@sum(EL(x1*,affine(terms(term(a,v))),x2*,affine(terms(term(b,v))),x3*)) -> 
											{ return debug("sum2", `in, `sum(EL(affine(terms(term(a+b,v))),x1*,x2*,x3*))); }
			in@sum(EL(x1*,affine(terms(constant(a))),x2*,affine(terms(constant(b))),x3*)) 	-> 
											{ return debug("sum3", `in, `sum(EL(x1,x2,x3,affine(terms(constant(a+b)))))); }
			in@sum(EL(y))	-> 				{ return debug("sum4", `in, `y);	}
			in@sum(EL(x1*,sum((x2*)),x3*)) -> 
											{ return debug("sum5", `in, `sum(EL(x1*,x2*,x3*))); }
			in@sum(EL(x1*,affine(terms(constant(0l))),x2*)) -> 
											{ return debug("sum6", `in, `sum(EL(x1*,x2*))); }
			in@sum(EL(affine(terms(constant(a))))) -> 
											{ return debug("sum7", `in, `affine(terms(constant(a)))); }
			in@sum(EL())-> 				{ return debug("sum8", `in, `affine(terms(constant(0)))); }


				
			in@prod(EL(y))						-> { return debug("prod0", `in, `y);	}
			in@prod(EL(x1*,affine(terms(constant(0l))),x2*)) 	-> { return debug("prod1", `in, `affine(terms(constant(0)))); }
			in@prod(EL(x1*,affine(terms(constant(1l))),x2*)) 	-> { return debug("prod2", `in, `prod(EL(x1,x2))); }

			//The following two rules can create prod(..., affine(terms()), ...) which later gets converted to prod(..., 0, ...) yielding incorrect results
			//Also , PROD[2, N, N] -> SUM[PROD[2, N], 2*N] is not a correct result
			//diabling the rules due to these reasons @Tomofumi
			/*in@prod(EL(p1*, c@affine(terms(constant(x))), p2*, affine(terms(x1*, term(a,b), x2*)), p3*)) -> 
										{ return debug("prod3a", `in, `sum(EL(prod(EL(c, affine(terms(x1*,x2*)), p1*, p2*, p3*)), affine(terms(term((x)*(a), b)))))); }
			in@prod(EL(p1*, affine(terms(x1*, term(a,b), x2*)), p2*, c@affine(terms(constant(x))), p3*)) -> 
										{ return debug("prod3b", `in, `sum(EL(prod(EL(c, affine(terms(x1*,x2*)), p1*, p2*, p3*)), affine(terms(term((x)*(a), b)))))); }*/

			//constrained version of the above rules for simple cases
			in@prod(EL(affine(terms(constant(x))), affine(terms(term(a,b))))) -> { return debug("prod3xa", `in, `affine(terms(term((x)*(a), b)))); }
			in@prod(EL(affine(terms(term(a,b))), affine(terms(constant(x))))) -> { return debug("prod3xb", `in, `affine(terms(term((x)*(a), b)))); }
			
			in@prod(EL(affine(terms(constant(x))), affine(terms(x1*, term(a,b), x2*)))) -> { return debug("prod3xc", `in, `sum(EL(affine(terms(term((x)*(a), b))), prod(EL(affine(terms(constant(x))), affine(terms(x1*,x2*))))))); }
			in@prod(EL(affine(terms(x1*, term(a,b), x2*)), affine(terms(constant(x))))) -> { return debug("prod3xd", `in, `sum(EL(affine(terms(term((x)*(a), b))), prod(EL(affine(terms(constant(x))), affine(terms(x1*,x2*))))))); }


			in@prod(EL(p1*, affine(terms(constant(x))), p2*, affine(terms(constant(y))), p3*))->
											{ return debug("prod4", `in, `prod(EL(affine(terms(constant((x)*(y)))), p1*, p2*, p3*))); }
			in@prod(EL())-> 				{ return debug("prod5", `in, `affine(terms(constant(0)))); }

			in@min(EL(a))										-> { return debug("min0", `in, `a); }
			in@min(EL())										-> { return debug("min1", `in, `affine(terms(constant(0)))); }
			in@min(EL(x1*, a, x2*, a, x3*))					-> { return debug("min2", `in, `min(EL(a, x1*, x2*, x3*))); }
			in@min(EL(x1, min(EL(x2*)),x3*))				-> { return debug("min3a", `in, `min(EL(x1, x2*, x3*))); }
			in@min(EL(x1*, min(EL(x2*)),x3))				-> { return debug("min3b", `in, `min(EL(x1*, x2*, x3))); }
			in@min(EL(x1*,affine(terms(constant(a))),x2*,affine(terms(constant(b))),x3*)) 	-> 
																	{ return debug("min4", `in, `min(EL(affine(terms(constant(Math_min(a,b)))),x1,x2,x3))); }
			//in@min(EL(x1*,term(a,v),x2*,term(b,v),x3*)) 		-> { return `min((term(Math_min(a,b),v),x1,x2,x3)); }
			//in@min(EL(x1*,min(x2*),x3*)) 						-> { return `min((x1,x2,x3)); }

			in@max(EL(a))										-> { return debug("max0", `in, `a); }
			in@max(EL())										-> { return debug("max1", `in, `affine(terms(constant(0)))); }
			in@max(EL(x1*, a, x2*, a, x3*))					-> { return debug("max2", `in, `max(EL(a, x1*, x2*, x3*))); }
			in@max(EL(x1, max(EL(x2*)),x3*))				-> { return debug("max3a", `in, `max(EL(x1, x2*, x3*))); }
			in@max(EL(x1*, max(EL(x2*)),x3))				-> { return debug("max3b", `in, `max(EL(x1*, x2*, x3))); }
			in@max(EL(x1*,affine(terms(constant(a))),x2*,affine(terms(constant(b))),x3*)) 	-> 
																	{ return debug("max4", `in, `max(EL(affine(terms(constant(Math_max(a,b)))),x1,x2,x3))); }
			//in@max(EL(x1*,term(a,v),x2*,term(b,v),x3*)) 		-> { return `max((term(Math_max(a,b),v),x1,x2,x3)); }
			//in@max(EL(term(a,v)))								-> { return `term(a,v); }
			//in@max(EL(x1*,min((x2*)),x3*))					-> { return `max((x1,x2,x3)); }
		} 
	}
 

/**

			
	affine(x1*,constant(a),x2*,constant(b),x3*) -> { return `affine(x1,x2,x3,constant(a+b)); }
				affine(x1*,constant(0),x2*) 				-> { return `affine(x1,x2); }

				sum(x1*,affine(a),x2*,affine(b),x3*) 		-> { return `sum(affine(a,b),x1*,x2*,x3*); }
				sum(affine(a)) 								-> { return `affine(a); }
				sum(x1*,term(a,v),x2*,term(b,v),x3*) 		-> { return `sum(term(a+b,v),x1,x2,x3); }
				sum(x1*,constant(a),x2*,constant(b),x3*) 	-> { return `sum(x1,x2,x3,constant(a+b)); }
				sum(y)										-> { return `y;	}
				sum(x1*,sum((x2*)),x3*) 					-> { return `sum(x1,x2,x3); }
				sum(x1*,constant(0),x2*) 					-> { return `sum(x1,x2); }
				sum((constant(a)))							-> { return `constant(a); }
				sum(())										-> { return `constant(0); }
				
				prod(y)										-> { return `y;	}
				prod(x1*,constant(0),x2*) 					-> { return `constant(0); }
				prod(x1*,constant(1),x2*) 					-> { return `prod((x1,x2)); }
				prod(constant(x), term(a,b)) 				-> { return `term((x)*(a), b); }
				prod(term(a,b), constant(x)) 				-> { return `term((x)*(a), b); }

				min(x1*,term(a,v),x2*,term(b,v),x3*) 		-> { return `min((term(Math_min(a,b),v),x1,x2,x3)); }
				min(x1*,constant(a),x2*,constant(b),x3) 	-> { return `min((constant(Math_min(a,b)),x1,x2,x3)); }
				min(a)										-> { return `a; }
				min(x1*,min(x2*),x3*) 						-> { return `min((x1,x2,x3)); }

				max(x1*,term(a,v),x2*,term(b,v),x3*) 		-> { return `max((term(Math_max(a,b),v),x1,x2,x3)); }
				max(x1*,constant(a),x2*,constant(b),x3) 	-> { return `max((constant(Math_max(a,b)),x1,x2,x3)); }
				max(term(a,v))								-> { return `term(a,v); }
				max(x1*,min((x2*)),x3*) 					-> { return `max((x1,x2,x3)); }

				mul(1,a)									-> { return `a;}
				mul(0,a)									-> { return `constant(0);}
				x@mul(c,mul(a,v))							-> { int r=(`a)*(`c); return `mul(r,v);}
				mul(c,term(a,v))							-> { int r=(`a)*(`c); return `term(r,v);}

				mul(c,constant(a))				-> { 
					int r=(`a)*(`c);
					return `constant(r); 
				}
			
				mul(c, sum((sum((x*, term(a,b), y*))))) -> {
						int ca = (`c)*(`a);
						return `sum((term(ca, b), mul(c, sum((x, y)))));
				}
				
				mul(c,sum((Y*,term(a,v),X*)))				-> { 
					int r=(`a)*(`c);
					return `sum((term((r),v),mul(c,sum((Y,X))))); 
				}

 
				mul(c,sum((Y*,constant(a),X*)))				-> { 
					int r=(`a)*(`c);
					return `sum((constant(r),mul(c,sum((Y,X))))); 
				}
				

				floord(sum((X1*,term(a,v),X2*)),c)		-> { 
					if(((`a)%(`c))==0) {
						int r=(`a)/(`c);
						return `sum((term((r),v),floord(sum((X1,X2)),c)));
					} 
				}

				x@ceild(term(v,a),b)		-> { 
					int r=(`v+(`b)-1)/(`b);
					//System.err.println("x@ceild(term(v,a),b) ->x="+`x.expr()+",r="+r);
					return `constant(r);
				}

				x@ceild(constant(v),b)		-> { 
					int r=(`v+(`b)-1)/(`b);
					//System.err.println("x@ceild(constant(v,a),b) ->x="+`x.expr()+",r="+r);
					return `constant(r);
				}
				x@ceild(a@sum(s),b)		-> { 
					IntExpressionAdapter r = new IntExpressionAdapter(`a.expr().simplify());
					//System.err.println("x@ceild(a@sum(s),b)->x="+`x.expr()+",r="+r.expr());
					return `ceild(r,b);
				}

				x@ceild(a@mul(s,f),b)		-> { 
					IntExpressionAdapter r = new IntExpressionAdapter(`a.expr().simplify());
					//System.err.println("x@ceild(a@mul(s,f),b)->x="+`x.expr()+",r="+r.expr());
					return `ceild(r,b);
				}

				x@floord(a,b)		-> { 
					IntExpressionAdapter r = new IntExpressionAdapter(`a.expr().simplify());
					return `floord(r,b);
				}

				x@mod(a,b)		-> { 
					IntExpressionAdapter r = new IntExpressionAdapter(`a.expr().simplify());
					return `mod(r,b);
				}

				x@mul(a,b)		-> { 
					IntExpressionAdapter r = new IntExpressionAdapter(`b.expr().simplify());
					IntExpressionAdapter nr = `mul(a,r);
					//System.err.println("x@mul(a,b)->x="+`x.expr()+",nr="+nr.expr());
					return nr;
				}

				floord(constant(a),b)		-> { 
					int r=(`a)/(`b);
					return `constant(r);
				}

				mod(sum((X1*,term(a,v),X2*)),c)		-> { 
					if(((`a)%(`c))==0) {
						int r=(`a)/(`c);
						return `sum((mod(sum((X1,X2)),c)));
					} 
				}

				//mod(sum((mul(a,X),mul(b,Y))),d) -> {	return `mod(sum((mul(Math_mod(a,d),X),mul(Math_mod(b,d),Y))),d);	}


				ceild(sum(()),c)						-> { return `constant(0); }

				mod(sum(()),c)							-> { return `constant(0); }
				floord(sum(()),c)						-> { return `constant(0); }
				ceild(sum(()),c)						-> { return `constant(0); }
				mul(c,sum(()))							-> { return `constant(0); }
		

**/
}